using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using InABox.Core;

namespace InABox.DynamicGrid;

public class FilterNode<T> : BaseFilterNode
{

    private Button ClearButton;
    private Button? LabelButton;
    private PropertyNode<T> Property;
    private OperatorNode Operator;
    private ValueNode Value;
    private Border ValuePlaceHolder;
    private ConstantNode Constant;
    private Button AndButton;
    private Button OrButton;
    private FilterNodeGrid<T> Ands;
    private FilterNodeGrid<T> Ors;

    private FilterNodeType filterType;

    public delegate void RemoveEventHandler();
    public event RemoveEventHandler? Remove;

    public FilterNode(Filter<T> filter, FilterNodeType filterType = FilterNodeType.ROOT)
    {
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(1, GridUnitType.Star) });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
        ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });

        RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto});
        RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto});
        RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto});

        ClearButton = new Button { Content = "-", Width = 20, Margin = new Thickness(0, 0, 5, 0), Padding = new Thickness(5) };

        if (filterType != FilterNodeType.ROOT)
        {
            ClearButton.Click += (s, e) => Remove?.Invoke();

            LabelButton = new Button { Content = "-", Width = 50, Margin = new Thickness(0, 0, 5, 0), Padding = new Thickness(5) };

            if (filterType == FilterNodeType.AND)
                LabelButton.Content = "And";
            else if (filterType == FilterNodeType.OR)
                LabelButton.Content = "Or";
        }
        else
        {
            ClearButton.Click += ClearButton_Click;
        }

        this.filterType = filterType;

        Property = new(filter.Expression) { Margin = new Thickness(0, 0, 5, 0) };
        Operator = new(filter.Operator) { Margin = new Thickness(0, 0, 5, 0) };

        ValuePlaceHolder = CreateValuePlaceHolder();
        
        var valueNode = Property.SelectedProperty != null ? CreateValueNode(Property.SelectedProperty, filter.Operator) : null;
        SetValueNode(valueNode ?? new EmptyValueNode(), filter.Value);

        Constant = CreateConstantNode(Property.SelectedProperty, filter.Operator, filter.Value as FilterConstant?);
        ConstantChanged(Constant.SelectedConstant);

        Ands = new FilterNodeGrid<T>(filter.Ands, FilterNodeType.AND) { Margin = new Thickness(0, 0, 0, 0) };
        Ors = new FilterNodeGrid<T>(filter.Ors, FilterNodeType.OR);

        AndButton = new Button { Content = "And", Width = 40, Margin = new Thickness(0, 0, 5, 0), Padding = new Thickness(5) };
        OrButton = new Button { Content = "Or", Width = 40, Margin = new Thickness(0, 0, 0, 0), Padding = new Thickness(5) };
        AndButton.Click += AndButton_Click;
        OrButton.Click += OrButton_Click;

        Property.PropertyChanged += Property_PropertyChanged;
        Operator.OperatorChanged += Operator_OperatorChanged;
        
        Add(ClearButton, 0, 0);
        if(LabelButton != null)
        {
            Add(LabelButton, 0, 1);
        }
        Add(Property, 0, 2);
        Add(Operator, 0, 3);
        Add(Constant, 0, 4);
        Add(ValuePlaceHolder, 0, 5);
        Add(AndButton, 0, 6);
        Add(OrButton, 0, 7);

        Add(Ands, 1, 1, 7);
        Add(Ors, 2, 1, 7);
    }

    private void ClearButton_Click(object sender, RoutedEventArgs e)
    {
        Property.Clear();
        Operator.Clear();
        Ands.Clear();
        Ors.Clear();
    }

    private void AndButton_Click(object sender, RoutedEventArgs e)
    {
        Ands.AddFilter(new Filter<T>());
    }

    private void OrButton_Click(object sender, RoutedEventArgs e)
    {
        Ors.AddFilter(new Filter<T>());
    }

    public override Filter<T>? GetFilter()
    {
        if (Operator.SelectedOperator == Core.Operator.All)
            return new Filter<T>().All();
        if (Operator.SelectedOperator == Core.Operator.None)
            return new Filter<T>().None();

        if (Property.SelectedProperty is null)
            return null;
        
        var filter = new Filter<T>();
        filter.Expression = CoreUtils.CreateMemberExpression(typeof(T), Property.SelectedProperty);
        filter.Operator = Operator.SelectedOperator;
        filter.Value = Constant.SelectedConstant ?? Value.Value;

        foreach(var and in Ands.GetFilters())
        {
            filter.Ands.Add(and);
        }

        foreach(var or in Ors.GetFilters())
        {
            filter.Ors.Add(or);
        }

        return filter;
    }

    private Border CreateValuePlaceHolder()
    {
        Border result = new Border()
        {
            BorderThickness = new Thickness(0.75),
            BorderBrush = Brushes.Gray,
            Background = Brushes.WhiteSmoke,
            Padding = new Thickness(0),
            Margin = new Thickness(0,0,5,0)
        };
        return result;
    }

    private Type? ValueNodeType(Type propertyType)
    {
        if (propertyType == typeof(bool))
            return typeof(BooleanValueNode);
        if (propertyType == typeof(string))
            return typeof(StringValueNode);
        if (propertyType == typeof(Guid))
            return typeof(GuidValueNode);
        if (propertyType.IsOrdinal())
            return typeof(IntegerValueNode);
        if (propertyType == typeof(double))
            return typeof(DoubleValueNode);
        if (propertyType == typeof(DateTime))
            return typeof(DateTimeValueNode);
        if (propertyType == typeof(TimeSpan))
            return typeof(TimeSpanValueNode);
        if (propertyType.GetTypeInfo().IsEnum)
            return typeof(EnumValueNode<>).MakeGenericType(propertyType);
        return null;
    }

    private ValueNode? CreateValueNode(Type propertyType, Operator op)
    {
        switch (op)
        {
            case Core.Operator.InList:
            case Core.Operator.NotInList:
                var elementNodeType = ValueNodeType(propertyType);
                if (elementNodeType == null)
                    return null;
                return Activator.CreateInstance(typeof(ArrayValueNode<>).MakeGenericType(elementNodeType)) as ValueNode;

            case Core.Operator.BeginsWith:
            case Core.Operator.Contains:
            case Core.Operator.EndsWith:
                return new StringValueNode();

            case Core.Operator.InQuery:
                return new SubQueryNode();

            case Core.Operator.All:
            case Core.Operator.None:
                return null;
        }
        var valueNodeType = ValueNodeType(propertyType);
        if (valueNodeType == null)
            return null;
        return Activator.CreateInstance(valueNodeType) as ValueNode;
    }

    private ValueNode? CreateValueNode(string propertyName, Operator op)
    {
        var property = CoreUtils.GetProperty(typeof(T), propertyName);
        return CreateValueNode(property.PropertyType, op);
    }

    [MemberNotNull(nameof(Value))]
    private void SetValueNode(ValueNode node, object? value)
    {
        Value = node;
        Value.Value = value;
        Value.Margin = new Thickness(0, 0, 5, 0);
        Add(Value, 0, 5);
        CheckValueVisibility();
    }
        
    private ConstantNode CreateConstantNode(string? propertyName, Operator op, FilterConstant? constantvalue)
    {
        var result = new ConstantNode();

        Dictionary<String, FilterConstant?> values = new Dictionary<string, FilterConstant?>();
        if (!String.IsNullOrWhiteSpace(propertyName))
        {
            var property = CoreUtils.GetProperty(typeof(T), propertyName);
            var constants = FilterConstants.Constants(property.PropertyType, op);
            if (constants.Any())
            {
                values[""] = null;
                foreach (var constant in constants)
                    values[CoreUtils.Neatify(constant.ToString())] = constant;
            }
        }

        result.DisplayMemberPath = "Key";
        result.SelectedValuePath = "Value";
        result.ItemsSource = values;
        result.SelectedConstant = constantvalue; 
        result.Margin = new Thickness(0, 0, 5, 0);
        result.Visibility = values.Any()
            ? Visibility.Visible
            : Visibility.Collapsed;
        result.ConstantChanged += ConstantChanged;
        return result;
    }
        
        
    private void Property_PropertyChanged(string? property)
    {
        Children.Remove(Constant);
        Constant = CreateConstantNode(property, Operator.SelectedOperator, null);
        Add(Constant,0,4);
            
        var value = Value.Value;
        Children.Remove(Value);
        var valueNode = property != null ?
            CreateValueNode(property, Operator.SelectedOperator) ?? new EmptyValueNode() :
            new EmptyValueNode();
        SetValueNode(
            valueNode,
            value
        );

    }

    private void Operator_OperatorChanged(Operator op)
    {

        Children.Remove(Constant);
        Constant = CreateConstantNode(Property.SelectedProperty, Operator.SelectedOperator, null);
        Add(Constant,0,4);
        
        var value = Value.Value;
        Children.Remove(Value);

        ValueNode node;
        if (Property.SelectedProperty == null)
            node = new EmptyValueNode();
        else
            node = CreateValueNode(Property.SelectedProperty, op) ?? new EmptyValueNode();
        SetValueNode(
            node,
            value
        );
    }

    private void ConstantChanged(FilterConstant? constant)
    {
        CheckValueVisibility();
    }

    private void CheckValueVisibility()
    {
        Value.Visibility = (Constant?.SelectedConstant == null) && Value is not EmptyValueNode
            ? Visibility.Visible
            : Visibility.Collapsed;
        ValuePlaceHolder.Visibility = (Constant?.SelectedConstant != null) || (Value is EmptyValueNode)
            ? Visibility.Visible
            : Visibility.Collapsed;
    }

    private void Add(UIElement element, int row, int column = 1, int columnSpan = 1)
    {
        element.SetValue(RowProperty, row);
        element.SetValue(ColumnProperty, column);
        element.SetValue(ColumnSpanProperty, columnSpan);
        Children.Add(element);
    }
}